''' Implements supported protocols for Kako simulations. '''

from kako.simulation.server import error

from kako.simulation.server import tcp
from kako.simulation.server import http
from kako.simulation.server import telnet



''' Implements Kako simulations. '''

from kako.simulation import manifest
from kako.simulation import server

from kako.simulation import http
from kako.simulation import https
from kako.simulation import telnet



import time
import socket
import logging
import binascii
import http.server
import socketserver

from kako import messaging


class RequestHandler(http.server.SimpleHTTPRequestHandler):
    ''' Implements HTTP handling for client connections. '''
    port = 'UNKNOWN'
    protocol = 'http'
    simulation = 'UNKNOWN'
    vulnerability = 'UNKNOWN'
    simulation_version = 'UNKNOWN'

    def __init__(self, request, client_address, server):
        ''' Extends parent with additional SNS channel and logger. '''
        self.log = logging.getLogger(__name__)
        self.port = server.manifest['port']
        self.version = server.manifest['version']
        self.protocol = server.manifest['protocol']
        self.simulation = server.manifest['name']
        self.simulation_version = server.manifest['version']

        self.banner = server.manifest['server']['banner']
        self.default_response = server.manifest['server']['response']

        self.routes_get = server.manifest['server']['routes']['get']
        self.routes_put = server.manifest['server']['routes']['put']
        self.routes_post = server.manifest['server']['routes']['post']
        self.routes_head = server.manifest['server']['routes']['head']
        self.routes_patch = server.manifest['server']['routes']['patch']
        self.routes_options = server.manifest['server']['routes']['options']

        # Suppress client errors.
        self.error_message_format = ''

        # Finish initialization using the parent class. This is actually more
        # like calling the great-grantparent, but hey!
        socketserver.BaseRequestHandler.__init__(
            self, request, client_address, server
        )

    def log_message(self, fmt, *args):
        ''' Override default STDx logger and write to Kako log handle. '''
        self.log.info(fmt % args)

    def log_error(self, fmt, *args):
        ''' Override default error logger and nerf messages. '''
        pass

    def log_request(self, code='-', size='-'):
        ''' Override default request logger and fix formatting. '''
        # Ensure both the HTTP path and command are set.
        try:
            command = self.command
        except AttributeError:
            command = 'NO_HTTP_COMMAND'
        try:
            path = self.path
        except AttributeError:
            path = 'NO_HTTP_PATH'

        # Log it.
        self.log_message(
            'Received HTTP %s for "%s" from %s. Responded with HTTP %s.',
            command,
            path,
            ':'.join(str(x) for x in self.client_address),
            str(code)
        )

    def version_string(self):
        ''' Overrides built-in version headers in HTTP responses.'''
        return self.banner

    def send_response(self, code, message=None):
        ''' Overrides built-in method to suppress headers if required. '''
        self.log_request(code)
        self.send_response_only(code, message)
        self.send_header('Date', self.date_time_string())

        # Only include a server header if its not empty.
        if self.version_string():
            self.send_header('Server', self.version_string())

    def do_generic(self, routes):
        ''' Implements generic route handling for requests. '''
        response = None
        if self.path is not None:
            for candidate in routes:
                if self.path.split('?')[0] == candidate['route']:
                    response = candidate['response']
                    self.vulnerability = candidate['vulnerability']
                    break

        # If no match was found, use the default response.
        if response is None:
            response = self.default_response

        # Capture the request.
        self.capture(
            self.rfile.read(
                int(self.headers.get('content-length', 0))
            )
        )

        # Reply, bolting on any applicable headers and response body.
        self.send_response(response['code'], response['text'])
        for header in response['headers']:
            self.send_header(header['key'], header['value'])
        self.end_headers()
        self.wfile.write(response['body'].encode())
        self.wfile.write(b'\r\n')

    def do_GET(self):
        ''' Implements routing for HTTP GET requests. '''
        self.do_generic(routes=self.routes_get)

    def do_PUT(self):
        ''' Implements routing for HTTP PUT requests. '''
        self.do_generic(routes=self.routes_put)

    def do_POST(self):
        ''' Implements routing for HTTP POST requests. '''
        self.do_generic(routes=self.routes_post)

    def do_HEAD(self):
        ''' Implements routing for HTTP HEAD requests. '''
        self.do_generic(routes=self.routes_head)

    def do_PATCH(self):
        ''' Implements routing for HTTP PATCH requests. '''
        self.do_generic(routes=self.routes_patch)

    def do_OPTIONS(self):
        ''' Implements routing for HTTP HEAD requests. '''
        self.do_generic(routes=self.routes_options)

    def capture(self, payload):
        ''' Implements 'capture' functionality for identified requests. '''
        # There doesn't appear to be a way to get the RAW HTTP input prior to
        # http.server having processed it. As a result, we need to recreate the
        # interaction(s) here.
        request = []
        request.append(self.raw_requestline.decode())
        for key, value in self.headers.items():
            request.append('{0}: {1}\r\n'.format(key, value))
        request.append('\r\n')

        # Re-add the payload.
        try:
            request.append(payload.decode())
        except UnicodeDecodeError:
            self.log.warning('Decoding payload failed, leaving as binary')
            request.append(binascii.hexlify(payload).decode())
        except (binascii.Error, binascii.Incomplete):
            self.log.error('Payload is totally broken, dropping!')
            request.append('MALFORMED PAYLOAD')

        # Send!
        msg = messaging.capture.Capture(
            timestamp=int(time.time()),
            capture=request,
            vulnerability=self.vulnerability,
            node=socket.gethostname(),
            destination_ip='TODO',
            destination_port=self.port,
            source_ip=self.client_address[0],
            source_port=self.client_address[1],
            simulation_name=self.simulation,
            simulation_version=self.simulation_version
        )

        # Publish to results queue.
        self.log.info('Publishing results from interaction to queue')
        self.server.results.put(msg.toJSON())



''' Implements an AWS SNS output / results processor for Kako. '''

import logging
import multiprocessing
import boto3

from botocore.exceptions import ClientError


class Processor(multiprocessing.Process):
    ''' AWS SNS output / results processor for Kako. '''

    def __init__(self, configuration, results, *args, **kwargs):
        super(Processor, self).__init__()

        self.log = logging.getLogger(__name__)
        self.results = results
        self.configuration = configuration

        # Bring down the logger level(s) for boto to prevent log spam during
        # AWS operations - such as pushing messages to SNS.
        logging.getLogger('boto3').setLevel(logging.WARNING)
        logging.getLogger('botocore').setLevel(logging.WARNING)

        # Setup an AWS SNS client for publishing captures.
        self.output = boto3.client(
            'sns',
            region_name=self.configuration['results']['attributes']['region']
        )

    def write(self, payload):
        ''' Implements a helper to write the provided payload to SNS. '''
        self.output.publish(
            TopicArn=self.configuration['results']['attributes']['topic'],
            Message=payload
        )

    def run(self):
        ''' Implements the main runable for the processor. '''
        self.log.info('Setting up AWS SNS results / output processor')
        while True:
            # When there is data in the queue, attempt to pull it and write to
            # the output topic. If something goes awry, requeue the message
            # prior to throwing the exception.
            if self.results.qsize() > 0:
                self.log.info(
                    '%s interaction captures in the queue',
                    self.results.qsize()
                )
                interaction = self.results.get()
                self.log.debug('Attempting to write interaction to SNS')
                try:
                    self.write(interaction)
                except (AttributeError, ClientError):
                    self.log.error('Requeuing interaction, as write failed...')
                    self.results.put(interaction)
                    raise
                self.log.debug('Interaction written okay')



''' Implements a file output / results processor for Kako. '''

import os
import logging
import multiprocessing

import json


class Processor(multiprocessing.Process):
    ''' File output / results processor for Kako. '''

    def __init__(self, configuration, results, *args, **kwargs):
        super(Processor, self).__init__()

        self.log = logging.getLogger(__name__)
        self.results = results
        self.configuration = configuration
        self.output = os.path.join(
            self.configuration['results']['attributes']['path'],
            'kako.json'
        )

    def write(self, payload):
        ''' Implements a helper to write the provided payload to file. '''
        with open(self.output, 'a') as hndl:
            hndl.write(payload)
            hndl.write('\r\n')

    def run(self):
        ''' Implements the main runable for the processor. '''
        self.log.info('Setting up File results / output processor')
        while True:
            # When there is data in the queue, attempt to pull it and write to
            # the output file. If something goes awry, requeue the message
            # prior to throwing the exception.
            if self.results.qsize() > 0:
                self.log.info(
                    '%s interaction captures in the queue',
                    self.results.qsize()
                )
                interaction = self.results.get()
                self.log.debug('Attempting to write interaction to file')
                try:
                    self.write(interaction)
                except (AttributeError, IOError, PermissionError):
                    self.log.error('Requeuing interaction, as write failed...')
                    self.results.put(interaction)
                    raise
                self.log.debug('Interaction written okay')



import time
import logging
import socket
import socketserver

from kako import messaging

from kako.simulation.server import error


class RequestHandler(socketserver.BaseRequestHandler):
    ''' Implements TCP handling for client connections. '''
    port = 'UNKNOWN'
    protocol = 'tcp'
    simulation = 'UNKNOWN'
    vulnerability = 'UNKNOWN'
    simulation_version = 'UNKNOWN'

    def __init__(self, request, client_address, server):
        ''' Extends parent with additional SNS channel and logger. '''
        self.log = logging.getLogger(__name__)
        self.port = server.manifest['port']
        self.version = server.manifest['version']
        self.protocol = server.manifest['protocol']
        self.simulation = server.manifest['name']
        self.simulation_version = server.manifest['version']

        # Results tracking.
        self.buffer = bytearray()
        self.record = bytearray()

        # Finish initialization using the parent class.
        socketserver.BaseRequestHandler.__init__(
            self, request, client_address, server
        )

    def write(self, message):
        ''' Writes to the socket. '''
        self.request.sendall(message)

    def read(self, length):
        ''' Reads from the socket into the record and build a byte-array. '''
        self.buffer = bytearray(self.request.recv(length))
        if not self.buffer:
            raise error.ClientDisconnect()

        # Save the read data into the record buffer.
        self.record.extend(self.buffer)

    def capture(self):
        ''' Implements 'capture' functionality for identified requests. '''
        msg = messaging.capture.Capture(
            timestamp=int(time.time()),
            capture=self.record.decode(errors='backslashreplace'),
            vulnerability=self.vulnerability,
            node=socket.gethostname(),
            destination_ip='TODO',
            destination_port=self.port,
            source_ip=self.client_address[0],
            source_port=self.client_address[1],
            simulation_name=self.simulation,
            simulation_version=self.simulation_version
        )

        # Publish to results queue.
        self.log.info('Publishing results from interaction to queue')
        self.server.results.put(msg.toJSON())


class Server(socketserver.ThreadingTCPServer):
    ''' Extends SocketServer ThreadingTCPServer to provide configurability. '''

    def __init__(self, server_address, RequestHandlerClass, manifest,
                 configuration, results):
        socketserver.ThreadingTCPServer.__init__(
            self, server_address, RequestHandlerClass
        )
        self.configuration = configuration
        self.manifest = manifest
        self.results = results



''' Implements manifest validation for Kako simulations. '''

from cerberus import Validator
from cerberus import schema_registry

# Define a generic schema for response entities.
SCHEMA_RESPONSE = {
    'code': {'type': 'integer'},
    'text': {'type': 'string'},
    'body': {'type': 'string'},
    'headers': {
        'type': 'list',
        'schema': {
            'type': 'dict',
            'schema': {
                'key': {'type': 'string'},
                'value': {'type': 'string'},
            }
        }
    }
}

# Define a generic schema for routing entities.
SCHEMA_ROUTING = {
    'type': 'dict',
    'schema': {
        'vulnerability': {'type': 'string'},
        'route': {'type': 'string'},
        'response': {
            'type': 'dict',
            'oneof_schema': [SCHEMA_RESPONSE]
        }
    }
}

# Define the configuration schema (for validation).
SCHEMA = {
    'name': {'type': 'string'},
    'port': {'type': 'integer'},
    'version': {'type': 'string'},
    'protocol': {'type': 'string'},
    'certificate': {'type': 'string'},
    'server': {
        'type': 'dict',
        'schema': {
            'banner': {'type': 'string'},
            'response': {
                'type': 'dict',
                'oneof_schema': [SCHEMA_RESPONSE]
            },
            'routes': {
                'type': 'dict',
                'schema': {
                    'patch': {
                        'type': 'list',
                        'oneof_schema': [SCHEMA_ROUTING]
                    },
                    'put': {
                        'type': 'list',
                        'oneof_schema': [SCHEMA_ROUTING]
                    },
                    'post': {
                        'type': 'list',
                        'oneof_schema': [SCHEMA_ROUTING]
                    },
                    'get': {
                        'type': 'list',
                        'oneof_schema': [SCHEMA_ROUTING]
                    },
                    'options': {
                        'type': 'list',
                        'oneof_schema': [SCHEMA_ROUTING]
                    },
                    'head': {
                        'type': 'list',
                        'oneof_schema': [SCHEMA_ROUTING]
                    },
                }
            }
        }
    }
}


def validate(config):
    ''' Test whether the provided configuration object is valid. '''
    linter = Validator(SCHEMA)
    if not linter.validate(config):
        raise AttributeError(linter.errors)



''' Implements interaction capture and serialization for Kako. '''

import json
import base64


class Capture(object):
    ''' Defines a capture message for submission of results from Kako. '''

    def __init__(self=None, destination_port=None, destination_ip=None,
                 source_ip=None, source_port=None, simulation_name=None,
                 capture=None, node=None, simulation_version=None,
                 timestamp=None, vulnerability=None):
        ''' Generates an object for this capture. '''
        self.node = node
        self.source_ip = source_ip
        self.timestamp = timestamp
        self.source_port = source_port
        self.vulnerability = vulnerability
        self.destination_ip = destination_ip
        self.simulation_name = simulation_name
        self.destination_port = destination_port
        self.simulation_version = simulation_version
        self.capture = base64.b64encode(''.join(capture).encode()).decode()

    def toJSON(self):
        ''' Provide a JSON serializer for this Object. '''
        return json.dumps(self.__dict__, sort_keys=True)



from setuptools import find_packages, setup, Command

# https://github.com/pypa/pypi-legacy/issues/148
try:
    import pypandoc
    LONG_DESCRIPTION = pypandoc.convert('README.md', 'rst')
except ImportError:
    LONG_DESCRIPTION = open('README.md').read()

setup(
    name='kako',
    version='1.1.0',
    description='Manifest driven IoT honeypots',
    long_description=LONG_DESCRIPTION,
    author='Peter Adkins',
    author_email='peter.adkins@kernelpicnic.net',
    url='https://www.github.com/darkarnium/kako',
    packages=find_packages('src'),
    license='MIT',
    download_url='https://github.com/darkarnium/kako/archive/1.1.0.tar.gz',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
    ],
    package_dir={
        'kako': 'src/kako',
    },
    scripts=[
        'src/kako-master'
    ],
    setup_requires=[
        'pytest-runner',
    ],
    tests_require=[
        'pytest',
        'pytest-cov',
    ],
    install_requires=[
        'click==7.1.2',
        'boto3==1.14.48',
        'PyYAML==5.3.1',
        'requests==2.24.0',
        'cerberus==1.3.2',
    ]
)



''' Implements configuration validation for Kako. '''

from cerberus import Validator

# Define a sub-schema for the result configuration for Amazon SNS.
SCHEMA_RESULTS_SNS = {
    'processor': {'type': 'string', 'allowed': ['sns']},
    'attributes': {
        'type': 'dict',
        'schema': {
            'region': {'type': 'string'},
            'topic': {'type': 'string'}
        }
    }
}

# Define a sub-schema for the result configuration for JSON.
SCHEMA_RESULTS_FILE = {
    'processor': {'type': 'string', 'allowed': ['file']},
    'attributes': {
        'type': 'dict',
        'schema': {
            'path': {'type': 'string'}
        }
    }
}

# Define the configuration schema (for validation).
SCHEMA = {
    'logging': {
        'type': 'dict',
        'schema': {
            'path': {'type': 'string'}
        }
    },
    'simulations': {
        'type': 'dict',
        'schema': {
            'path': {'type': 'string'}
        }
    },
    'monitoring': {
        'type': 'dict',
        'schema': {
            'enabled': {'type': 'boolean'},
            'interval': {'type': 'integer'}
        }
    },
    'results': {
        'type': 'dict',
        'oneof_schema': [
            SCHEMA_RESULTS_SNS,
            SCHEMA_RESULTS_FILE
        ]
    }
}


def validate(config):
    ''' Test whether the provided configuration object is valid. '''
    linter = Validator(SCHEMA)
    if not linter.validate(config):
        raise AttributeError(linter.errors)



''' Implements manifest driven Telnet simulation for Kako. '''

import logging
import multiprocessing

from kako.simulation.server import tcp
from kako.simulation.server import telnet


class Simulation(multiprocessing.Process):
    ''' Simulation for a vulnerable Telnet service. '''

    def __init__(self, manifest, configuration, results, *args, **kwargs):
        super(Simulation, self).__init__()

        self.log = logging.getLogger(__name__)
        self.results = results
        self.manifest = manifest
        self.configuration = configuration

    def run(self):
        ''' Implements the main runable for the simulation. '''
        self.log.info(
            'Setting up listener on TCP/%s', str(self.manifest['port'])
        )
        service = tcp.Server(
            ('0.0.0.0', self.manifest['port']),
            telnet.RequestHandler,
            self.manifest,
            self.configuration,
            self.results
        )
        service.serve_forever()



''' Implements Linux command-line (Busybox) emulation for Kako. '''

import re
import binascii

from kako.simulation.server import error


class CommandInterpreter(object):
    ''' Implements commands for an embedded BusyBox system (Linux). '''
    version = 'BusyBox v1.19.3 (2013-11-01 10:10:26 CST)'

    def do_id(self, args=None):
        ''' Returns `id` command output. '''
        if args is None:
            args = []
        return 'uid=0(root) gid=0(root) groups=0(root)'

    def do_exit(self, args=None):
        ''' Returns an exit signal to the caller. '''
        if args is None:
            args = []
        raise error.ClientCommandExit()

    def do_ls(self, args=None):
        ''' Stub `ls`. '''
        if args is None:
            args = []
        return

    def do_rm(self, args=None):
        ''' Stub `rm`. '''
        if args is None:
            args = []
        return

    def do_cp(self, args=None):
        ''' Stub `cp`. '''
        if args is None:
            args = []
        return

    def do_cd(self, args=None):
        ''' Stub `cd`. '''
        if args is None:
            args = []
        return

    def do_wget(self, args=None):
        ''' Stub `wget`. '''
        if args is None:
            args = []
        return

    def do_busybox(self, args=None):
        ''' Wrapper the handle() method to handle BusyBox applets. '''
        if args:
            result = self.handle(' '.join(args))
            if re.match(r'^sh:\s', result):
                return '{0}: applet not found'.format(args[0])
            else:
                return result
        else:
            return self.version

    def do_echo(self, args=None):
        ''' Returns the input as provided - sans 'arguments' '''
        if args is None:
            args = []

        # Remove any 'arguments' from echo only if contiguous and before any
        # other data.
        for idx, arg in enumerate(args):
            if arg.startswith('-'):
                del args[idx]
            else:
                break

        # Decode encoded characters.
        decoded_args = []
        for arg in args:
            if "\\x" in arg:
                try:
                    decoded_args.append(
                        re.sub(
                            r'\\x([0-9A-Z]{1,2})',
                            lambda x: binascii.unhexlify(x.group(1)).decode(),
                            arg,
                            flags=re.IGNORECASE
                        )
                    )
                except UnicodeDecodeError as _:
                    decoded_args.append(arg)
            else:
                decoded_args.append(arg)

        # Flag for quotes to be stripped, if they're matched.
        matched_double_quote = False
        matched_single_quote = False
        if not ''.join(decoded_args).count('"') % 2:
            matched_double_quote = True
        if not ''.join(decoded_args).count("'") % 2:
            matched_single_quote = True

        # Strip all quotes, if required, and return it.
        result = ' '.join(decoded_args).strip()
        result = result.replace('"', '') if matched_double_quote else result
        result = result.replace("'", '') if matched_single_quote else result
        return result.strip()

    def do_cat(self, args=None):
        ''' Implement expected Mirai files. '''
        if args is None:
            args = []
        if len(args) < 1:
            return

        # TELNET_DETECT_ARCH - Fake ELF header.
        if args[0] == '/bin/echo':
            return ''.join([
                '\x7F\x45\x4c\x46',                  # ELF magic.
                '\x01',                              # 32-Bit.
                '\x01',                              # Little Endian.
                '\x01',                              # Version 1.
                '\x03',                              # Linux ABI.
                '\x00\x00\x00\x00\x00\x00\x00\x00',  # Unused padding.
                '\x02',                              # Executable.
                '\x08'                               # MIPS.
            ])

        # TELNET_PARSE_MOUNTS - Fake read-write mount-point(s).
        if args[0] == '/proc/mounts':
            return 'rootfs / rootfs rw 0 0'

    def do_ps(self, args=None):
        ''' Implements Mirai expected processes. '''
        if args is None:
            args = []
        process_list = []
        process_list.append('PID   Uid    VmSize    Stat    Command\r\n')
        process_list.append('  1   root      404     S      init')
        return ''.join(process_list)

    def handle(self, commands):
        ''' Dispatches the input command to the relevant handler. '''
        # Split commands on semicolon, in the case there are multiple commands
        # on one line.
        output = ''
        for command in commands.split(';'):
            # Split command from arguments, and ensure we're left with at least
            # a command to execute.
            arguments = re.split(r'\s+', command.lstrip())
            if len(arguments) < 1:
                continue

            # Remove full path from command - if present - and attempt to call.
            command = arguments[0].split('/')[-1]
            arguments.pop(0)
            try:
                ref = getattr(self, "do_{0}".format(command))
            except AttributeError:
                if len(command) >= 1:
                    output += "sh: {0}: command not found\r\n".format(command)
                continue

            # Call the appropriate handler, and record the output.
            result = ref(arguments)
            if result:
                output += '{0}\r\n'.format(result.strip())
            else:
                output += '\r\n'

        return output



import logging
import pprint

from kako.simulation.server import tcp
from kako.simulation.server import error
from kako.simulation.system import linux


class RequestHandler(tcp.RequestHandler):
    ''' Implements Telnet handling for client connections. '''
    port = 'UNKNOWN'
    protocol = 'tcp'
    simulation = 'UNKNOWN'
    vulnerability = 'UNKNOWN'
    simulation_version = 'UNKNOWN'

    def __init__(self, request, client_address, server):
        ''' Setup versions and system names for this service. '''
        self.prompt = '#'
        self.hostname = 'default'

        self.banner = server.manifest['server']['banner']
        self.port = server.manifest['port']
        self.version = server.manifest['version']
        self.protocol = server.manifest['protocol']
        self.simulation = server.manifest['name']
        self.simulation_version = server.manifest['version']

        self.interpreter = linux.CommandInterpreter()

        tcp.RequestHandler.__init__(self, request, client_address, server)

    def read(self, length=None):
        ''' Override base read() method with one that handles Telnet IACs. '''
        part_index = 0
        part_buffer = bytearray()
        while True:
            # Read byte-by-byte so that IACs can be handled.
            tcp.RequestHandler.read(self, 1)
            if self.buffer[0] == 255:
                tcp.RequestHandler.read(self, 1)
                # Reply to WILL with WON'T.
                if self.buffer[0] == 251:
                    tcp.RequestHandler.read(self, 1)
                    self.write(bytearray([255, 252, self.buffer[0]]))
                    continue
                # Reply to DO with DON'T.
                if self.buffer[0] == 253:
                    tcp.RequestHandler.read(self, 1)
                    self.write(bytearray([255, 254, self.buffer[0]]))
                    continue
                # If IACs sub-negotiation, process it (...by ignoring it)
                if self.buffer[0] == 250:
                    while self.buffer[0] != 240:
                        tcp.RequestHandler.read(self, 1)
                    continue

            # If it dropped through it's just data so keep it.
            part_index += 1
            part_buffer.extend(self.buffer)

            # If line-ending or the length constraint has been satisfied, bail.
            if self.buffer[0] == 10 or (length and part_index >= length):
                break

        # Fix the buffer to contain the FULL read - sans IACs.
        self.buffer = part_buffer

    def do_iacs(self):
        ''' Sends Telnet negotiation messages to the client (IAC). '''
        # Attempt at RFC854 and RFC1073 compliance.
        parameters = []
        parameters.append([255, 254, 1])   # IAC DON'T ECHO
        parameters.append([255, 254, 31])  # IAC DON'T NAWS

        # Write all negotiation messages to the client.
        for parameter in parameters:
            self.write(bytearray(parameter))

    def do_login(self):
        ''' Simulates a login prompt and captures the credentials. '''
        self.write(self.banner.encode())
        self.write(b'\r\n\r\n')
        self.write('{0} login: '.format(self.hostname).encode())
        self.read()
        self.write(b'Password: ')
        self.read()

    def do_motd(self):
        ''' Simulates an MOTD style banner (after login). '''
        self.write(b'\r\n')

    def handle(self):
        ''' Extend TCP RequestHandler to implement a fake telnet service. '''
        self.log.info(
            'Received connection from %s',
            ':'.join(str(x) for x in self.client_address)
        )

        # Negotiate with the client, and handle initial login / banner.
        try:
            self.do_iacs()
            self.do_login()
            self.do_motd()
        except error.ClientDisconnect as _:
            self.capture()
            self.log.info('Client force disconnected.')
            return

        # Kick off the REPL, and loop until do_repl returns False.
        self.log.info('Negotiation and login complete, dropping to "shell"')
        while True:
            self.write('{0} '.format(self.prompt.rstrip()).encode())
            try:
                self.read()
            except error.ClientDisconnect as _:
                self.log.info('Client force disconnected!')
                break

            # Convert read byte array into a string and process.
            cmd = ''.join(map(chr, self.buffer)).strip()
            try:
                self.write(self.interpreter.handle(cmd).encode())
            except error.ClientCommandExit as _:
                break

        # Record the interaction.
        self.capture()



class ClientDisconnect(Exception):
    ''' Exception for signalling client disconnect. '''
    pass


class ClientCommandExit(Exception):
    ''' Exception for signalling client requested 'exit'. '''
    pass


class ClientCommandNotFound(Exception):
    ''' Exception for signalling command not found. '''
    pass
